\set ECHO none

--
-- quick checks of functions and operators
--

-------------------------------------------------------------------------------
-- Sample Data
-------------------------------------------------------------------------------

CREATE TABLE one
(
  rand real NOT NULL,
  norm uncertain NOT NULL,
  hist uncertain NOT NULL,
  dist uncertain NOT NULL
);

--
-- x = seq(-2.75, 2.75, length=12)
-- data.frame(x, dnorm(x, mean=0, sd=1))
--
INSERT INTO one
  SELECT
    random(), uval, u_norm2hist(uval, 12), u_norm2dist(uval, 12)
  FROM (
    SELECT uncertain_parse('c', 'norm(0,1)') AS uval
  ) sub;

SELECT u_real2hist(3.4);

-------------------------------------------------------------------------------
-- Inquiry Functions
-------------------------------------------------------------------------------

SELECT
  u_lower(norm, 0) AS norm,
  u_lower(hist, 0) AS hist,
  u_lower(dist, 0) AS dist
FROM one;

SELECT
  u_upper(norm, 0) AS norm,
  u_upper(hist, 0) AS hist,
  u_upper(dist, 0) AS dist
FROM one;

-------------------------------------------------------------------------------
-- u_pdf
-------------------------------------------------------------------------------

SELECT
  u_pdf(norm, ARRAY[-3.25]) AS "pdf: -3.25",
  u_pdf(norm, ARRAY[-3.00]) AS "pdf: -3.00",
  u_pdf(norm, ARRAY[-2.75]) AS "pdf: -2.75",
  u_pdf(norm, ARRAY[1.25]) AS "pdf: 1.25",
  u_pdf(norm, ARRAY[1.50]) AS "pdf: 1.50",
  u_pdf(norm, ARRAY[2.75]) AS "pdf: 2.75",
  u_pdf(norm, ARRAY[3.00]) AS "pdf: 3.00",
  u_pdf(norm, ARRAY[3.25]) AS "pdf: 3.25"
FROM one;

SELECT
  u_pdf(hist, ARRAY[-3.25]) AS "pdf: -3.25",
  u_pdf(hist, ARRAY[-3.00]) AS "pdf: -3.00",
  u_pdf(hist, ARRAY[1.49]) AS "pdf: 1.49",
  u_pdf(hist, ARRAY[1.50]) AS "pdf: 1.50",
  u_pdf(hist, ARRAY[1.51]) AS "pdf: 1.51",
  u_pdf(hist, ARRAY[3.00]) AS "pdf: 3.00",
  u_pdf(hist, ARRAY[3.25]) AS "pdf: 3.25"
FROM one;

SELECT
  u_pdf(dist, ARRAY[-3.00]) AS "pdf: -3.00",
  u_pdf(dist, ARRAY[-2.75]) AS "pdf: -2.75",
  u_pdf(dist, ARRAY[1.24]) AS "pdf: 1.24",
  u_pdf(dist, ARRAY[1.25]) AS "pdf: 1.25",
  u_pdf(dist, ARRAY[1.26]) AS "pdf: 1.26",
  u_pdf(dist, ARRAY[2.75]) AS "pdf: 2.75",
  u_pdf(dist, ARRAY[3.00]) AS "pdf: 3.00"
FROM one;

-------------------------------------------------------------------------------
-- u_cdf
-------------------------------------------------------------------------------

SELECT
  u_cdf(norm, ARRAY[-3.25]) AS "cdf: -3.25",
  u_cdf(norm, ARRAY[-3.00]) AS "cdf: -3.00",
  u_cdf(norm, ARRAY[-2.75]) AS "cdf: -2.75",
  u_cdf(norm, ARRAY[1.25]) AS "cdf: 1.25",
  u_cdf(norm, ARRAY[1.50]) AS "cdf: 1.50",
  u_cdf(norm, ARRAY[2.75]) AS "cdf: 2.75",
  u_cdf(norm, ARRAY[3.00]) AS "cdf: 3.00",
  u_cdf(norm, ARRAY[3.25]) AS "cdf: 3.25"
FROM one;

SELECT
  u_cdf(hist, ARRAY[-3.25]) AS "cdf: -3.25",
  u_cdf(hist, ARRAY[-3.00]) AS "cdf: -3.00",
  u_cdf(hist, ARRAY[1.49]) AS "cdf: 1.49",
  u_cdf(hist, ARRAY[1.50]) AS "cdf: 1.50",
  u_cdf(hist, ARRAY[1.51]) AS "cdf: 1.51",
  u_cdf(hist, ARRAY[3.00]) AS "cdf: 3.00",
  u_cdf(hist, ARRAY[3.25]) AS "cdf: 3.25"
FROM one;

SELECT
  u_cdf(dist, ARRAY[-3.00]) AS "cdf: -3.00",
  u_cdf(dist, ARRAY[-2.75]) AS "cdf: -2.75",
  u_cdf(dist, ARRAY[1.24]) AS "cdf: 1.24",
  u_cdf(dist, ARRAY[1.25]) AS "cdf: 1.25",
  u_cdf(dist, ARRAY[1.26]) AS "cdf: 1.26",
  u_cdf(dist, ARRAY[2.75]) AS "cdf: 2.75",
  u_cdf(dist, ARRAY[3.00]) AS "cdf: 3.00"
FROM one;

-------------------------------------------------------------------------------
-- Range Queries
-------------------------------------------------------------------------------

SELECT
  u_prob(norm, ARRAY[-1.25], ARRAY[1.25]) AS norm,
  u_prob(hist, ARRAY[-1.25], ARRAY[1.25]) AS hist,
  u_prob(dist, ARRAY[-1.25], ARRAY[1.25]) AS dist
FROM one;

-------------------------------------------------------------------------------
-- Clean Up
-------------------------------------------------------------------------------

DROP TABLE one;
