-------------------------------------------------------------------------------
-- dst-funcs.sql : functions that generate random samples of distributions
-------------------------------------------------------------------------------

DROP FUNCTION IF EXISTS rnorm(_mean real, _sdev real);
DROP FUNCTION IF EXISTS runif(_minv real, _maxv real);
DROP FUNCTION IF EXISTS rexpo(_rate real);

-------------------------------------------------------------------------------

CREATE FUNCTION rnorm(_mean real, _sdev real)
RETURNS real AS $$
DECLARE
  x1 real;
  x2 real;
  w real;
  g real;
BEGIN

  -- like Java, using the polar form of the Box-Muller transformation
  -- see http://www.taygeta.com/random/gaussian.html for details
  w := 0;
  WHILE (w = 0 OR w >= 1.0) LOOP
    x1 := 2.0 * random() - 1.0;
    x2 := 2.0 * random() - 1.0;
    w := x1 * x1 + x2 * x2;
  END LOOP;

  -- unlike Java, we don't cache x2 for the subsequent call
  g := x1 * sqrt( (-2.0 * ln( w )) / w );
  RETURN $1 + $2 * g;

END;
$$ LANGUAGE plpgsql VOLATILE STRICT;

	COMMENT ON FUNCTION rnorm(_mean real, _sdev real)
	IS 'random sample from a normal distribution';

-------------------------------------------------------------------------------

CREATE FUNCTION runif(_minv real, _maxv real)
RETURNS real AS $$

  SELECT ( random() * ($2 - $1) + $1 )::real;

$$ LANGUAGE SQL VOLATILE STRICT;

	COMMENT ON FUNCTION runif(_minv real, _maxv real)
	IS 'random sample from a uniform distribution';

-------------------------------------------------------------------------------

CREATE FUNCTION rexpo(_rate real)
RETURNS real AS $$

  SELECT ( -1 * $1 * ln(random()) )::real;

$$ LANGUAGE SQL VOLATILE STRICT;

	COMMENT ON FUNCTION rexpo(_rate real)
	IS 'random sample from an exponential distribution';
