-------------------------------------------------------------------------------
-- loc-funcs.sql : functions that generate random locations (e.g. from sensors)
-------------------------------------------------------------------------------

DROP TYPE IF EXISTS location CASCADE;

CREATE TYPE location AS
(
  ts time, -- time stamp
  xpos integer, -- x coordinate
  ypos integer, -- y coordinate
  act integer -- activity level
);

	COMMENT ON TYPE location
	IS 'sensor time, coordinates, and activity';

-------------------------------------------------------------------------------

CREATE FUNCTION random_walk(INOUT location, integer)
RETURNS SETOF location AS $$
DECLARE
  -- initial activity influences total distance and time
  wild INTEGER := $1.act;
BEGIN

  -- generate the next result row, one by one
  FOR i IN 1..$2 LOOP
    $1.ts := $1.ts + rexpo(wild)::text::interval;
    $1.xpos := $1.xpos + rnorm(0,$1.act)::integer;
    $1.ypos := $1.ypos + rnorm(0,$1.act)::integer;
    $1.act := runif(1, wild * 2)::integer;
    RETURN NEXT;
  END LOOP;
  RETURN;

END;
$$ LANGUAGE plpgsql VOLATILE STRICT;

	COMMENT ON FUNCTION random_walk(INOUT location, integer)
	IS 'generates a random walk from a starting location';

-------------------------------------------------------------------------------

CREATE FUNCTION location_pdf(location)
RETURNS uncertain AS $$

  -- simple product of two gaussians centered on the points
  -- with standard deviation based on the activity level
  SELECT u_product(
    uncertain_parse( 'c' , ('norm(' || $1.xpos || ',' || $1.act || ')')::cstring ) ,
    uncertain_parse( 'c' , ('norm(' || $1.ypos || ',' || $1.act || ')')::cstring ) );

$$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION location_pdf(location)
	IS 'gets an uncertain version of the x,y location';
