/*
 * orion_bki.c - Orion API for the PostgreSQL backend
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "orion_bki.h"
#include "executor/spi.h"
#include "utils/lsyscache.h"
#include "utils/syscache.h"

/*---------------------------------------------------------------------------*\
  Extension of the internal function table.
  This list was generated using: grep -r "(PG_FUNCTION_ARGS);"
\*---------------------------------------------------------------------------*/

Datum pti_consistent(PG_FUNCTION_ARGS);
Datum pti_union(PG_FUNCTION_ARGS);
Datum pti_compress(PG_FUNCTION_ARGS);
Datum pti_decompress(PG_FUNCTION_ARGS);
Datum pti_penalty(PG_FUNCTION_ARGS);
Datum pti_picksplit(PG_FUNCTION_ARGS);
Datum pti_same(PG_FUNCTION_ARGS);

Datum u_floor(PG_FUNCTION_ARGS);
Datum u_floor_ex(PG_FUNCTION_ARGS);
Datum u_margin(PG_FUNCTION_ARGS);
Datum u_product(PG_FUNCTION_ARGS);
Datum u_lower(PG_FUNCTION_ARGS);
Datum u_upper(PG_FUNCTION_ARGS);
Datum u_pdf(PG_FUNCTION_ARGS);
Datum u_cdf(PG_FUNCTION_ARGS);
Datum u_prob(PG_FUNCTION_ARGS);
Datum u_mass(PG_FUNCTION_ARGS);
Datum u_real2hist(PG_FUNCTION_ARGS);
Datum u_sample(PG_FUNCTION_ARGS);
Datum u_reduce(PG_FUNCTION_ARGS);

Datum uncertain_in(PG_FUNCTION_ARGS);
Datum uncertain_parse(PG_FUNCTION_ARGS);
Datum uncertain_out(PG_FUNCTION_ARGS);
Datum uncertain_print(PG_FUNCTION_ARGS);

FmgrBuiltin orion_builtins[] = {
	{ 1, "pti_consistent", 3, false, false, pti_consistent },
	{ 1, "pti_union", 2, false, false, pti_union },
	{ 1, "pti_compress", 1, false, false, pti_compress },
	{ 1, "pti_decompress", 1, false, false, pti_decompress },
	{ 1, "pti_penalty", 3, true, false, pti_penalty },
	{ 1, "pti_picksplit", 2, false, false, pti_picksplit },
	{ 1, "pti_same", 3, false, false, pti_same },

	{ 1, "u_floor", 2, true, false, u_floor },
	{ 1, "u_floor_ex", 3, true, false, u_floor_ex },
	{ 1, "u_margin", 2, true, false, u_margin },
	{ 1, "u_product", 2, true, false, u_product },
	{ 1, "u_lower", 2, true, false, u_lower },
	{ 1, "u_upper", 2, true, false, u_upper },
	{ 1, "u_pdf", 2, true, false, u_pdf },
	{ 1, "u_cdf", 2, true, false, u_cdf },
	{ 1, "u_prob", 3, true, false, u_prob },
	{ 1, "u_mass", 1, true, false, u_mass },
	{ 1, "u_real2hist", 1, true, false, u_real2hist },
	{ 1, "u_sample", 2, true, false, u_sample },
	{ 1, "u_reduce", 2, true, false, u_reduce },

	{ 1, "uncertain_in", 1, true, false, uncertain_in },
	{ 1, "uncertain_parse", 2, true, false, uncertain_parse },
	{ 1, "uncertain_out", 1, true, false, uncertain_out },
	{ 1, "uncertain_print", 1, true, false, uncertain_print },
};

const int orion_nbuiltins = sizeof(orion_builtins) / sizeof(FmgrBuiltin);

/*---------------------------------------------------------------------------*\
  Initialization
\*---------------------------------------------------------------------------*/

/*
 * Uses SPI to get the actual oids for built-in functions. SearchSysCache in
 * PROCNAMEARGSNSP might be better, but we don't have oids for all the args.
 */
static void orion_init(Oid orion_schema) {
	char sql[100];
	TupleDesc tdesc;
	HeapTuple tuple;
	int32 i, j;
	char *name;
	bool isnull;

	sprintf(sql, "SELECT oid, proname FROM pg_proc WHERE pronamespace = %d",
			orion_schema);
	SPI_connect();
	if (SPI_execute(sql, true, 0) == SPI_OK_SELECT) {
		tdesc = SPI_tuptable->tupdesc;
		for (i = 0; i < SPI_processed; i++) {
			tuple = SPI_tuptable->vals[i];
			name = SPI_getvalue(tuple, tdesc, 2);

			/* scan the builtins table for the current name */
			for (j = 0; j < orion_nbuiltins; j++) {
				if (strcmp(name, orion_builtins[j].funcName) == 0) {
					orion_builtins[j].foid = DatumGetObjectId(
							SPI_getbinval(tuple, tdesc, 1, &isnull));
					break;
				}
			}
		}
	}
	SPI_finish();

} /* orion_init */

/*
 * Is Orion installed in this database?
 */
bool orion_present(void) {
	static bool first_call = true;
	static bool is_present = false;
	Oid orion_schema;
	Oid orion_table;

	/* check for Orion during the first call */
	if (first_call) {
		first_call = false;
		orion_schema = GetSysCacheOid(NAMESPACENAME, CStringGetDatum("orion"),
				0, 0, 0);
		if (orion_schema != InvalidOid) {
			orion_table = get_relname_relid("orion_present", orion_schema);
			if (orion_table != InvalidOid) {
				orion_init(orion_schema);
				is_present = true;
			}
		}
	}
	return is_present;
}
