/*
 * pdfs.c - catalog of supported PDF types
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "pdf.h"

/*---------------------------------------------------------------------------*\
  PDF Type Array
\*---------------------------------------------------------------------------*/

extern const PDF_vtable
  Binomial_vtbl,  /* discrete binomial distribution */
  Delta_vtbl,     /* distribution representing a certain value */
  Histogram_vtbl, /* generic multidimensional histogram */
	Histogramd_vtbl,/* same as multidimensional histogram, but parses densities */
  Normal_vtbl,    /* 1D categorical distribution */
  Poisson_vtbl,   /* discrete poisson distribution */
  Zero_vtbl;      /* N-dimensional zero probability distribution */

static const PDF_vtable *PDF_types[] = {
  &Binomial_vtbl,
  &Delta_vtbl,
  &Histogram_vtbl,
  &Histogramd_vtbl,
  &Normal_vtbl,
  &Poisson_vtbl,
  &Zero_vtbl,
  NULL};

/*---------------------------------------------------------------------------*\
  Lookup Functions
\*---------------------------------------------------------------------------*/

const PDF_vtable *pdf_name_to_vtbl(char *name) {
	int i=0;
	while(PDF_types[i] && strcmp(name, PDF_types[i]->name) != 0) ++i;
	return PDF_types[i];
}

const PDF_vtable *pdf_id_to_vtbl(char id) {
	int i=0;
	while (PDF_types[i] && PDF_types[i]->id != id) ++i;
	return PDF_types[i];
}

/*---------------------------------------------------------------------------*\
  Utility Functions
\*---------------------------------------------------------------------------*/

int ipow(int base, int exp) {
    int ret=1;
    while(exp--)
        ret *= base;
    return ret;
}
/*
float PDF_mass(PDF *pdf) {
	float upper[pdf->ndim];
	int i;
	for(i=0; i < pdf->ndim; ++i)
		upper[i] = PDF_upper(pdf, i);
	return PDF_cdf(pdf, upper);
}
*/