/*
 * zero.c - Zero-probability distribution
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "pdf.h"
#include "pdfs_io.h"
#include "uncertain.h"
#include "errors.h"
#include <math.h>

/*---------------------------------------------------------------------------*\
	Prototypes & VTable
\*---------------------------------------------------------------------------*/

typedef struct Zero {
	PDF     pdf;
} Zero;


static float4 zero_bound(PDF*, int);
static float4 zero_xdf(PDF*, float4*);
static float4 zero_mass(PDF*);
static PDF*   zero_sample(PDF*, int n);
static PDF*   zero_parse(char*, char*);
static char*  zero_print(PDF*);

#define PDF_NAME "zero"

const PDF_vtable Zero_vtbl = {
	.name = PDF_NAME,
	.id = '0',

	.lower = zero_bound,
	.upper = zero_bound,
	.pdf = zero_xdf,
	.cdf = zero_xdf,
	.mass = zero_mass,
	.sample = zero_sample,

	.parse = zero_parse,
	.print = zero_print
};

PDF *zero_init(int ndim) {
	PDF *pdf = palloc(sizeof(Zero));
	pdf->vtbl = &Zero_vtbl;
	pdf->size = sizeof(Zero);
	pdf->ndim = ndim;

	return pdf;
}

static float4 zero_bound(PDF *_pdf, int dim) {
	return 0.;
}

static float4 zero_xdf(PDF *_pdf, float4 *x) {
	return 0.;
}

static float4 zero_mass(PDF *_pdf) {
	return 0.;
}

static PDF *zero_sample(PDF *pdf, int n) {
	float4 bins[1] = {0.};
	HistHead hdr = {.minv = 0., .maxv = 0., .count = 1, .type = DIM_UDIS };

	return (PDF *) histogram_init(1, &hdr, bins);
}

/*---------------------------------------------------------------------------*\
	I/O
\*---------------------------------------------------------------------------*/

static PDF *zero_parse(char *str, char *dimtypes) {
	int32 ndim, i;
	char *pos=str;

	skipSpace(pos);
	pos = scanInt32(&ndim, pos, PDF_NAME, str);
	checkExtra(pos, str);

	for(i=0; i < ndim; ++i)
		dimtypes[i] = 'd';

	return zero_init(ndim);
}

static char *zero_print(PDF *pdf) {
	char *str = palloc(14);
	sprintf(str, "%i", pdf->ndim);
	return str;
}
