SET search_path TO orion;

-- Composite type for value threshold pair
CREATE TYPE u_pair AS (
    value   	real,
    threshold   real
);

-------------------------------------------------------------------------------------------
-- (Uncertain > real) functions
-------------------------------------------------------------------------------------------
-- Probability operator function for >
CREATE OR REPLACE FUNCTION u_gt_prob(uncertain, real) RETURNS real AS $$
DECLARE
    attr ALIAS FOR $1;
    value ALIAS FOR $2;
    temp_arr real[];
BEGIN
	temp_arr[0] := value;
    RETURN u_prob(u_margin(attr, '{0}'), temp_arr, '{inf}');
END;
$$ LANGUAGE plpgsql;

-- Probability operator for >
CREATE OPERATOR >% (
	PROCEDURE = u_gt_prob,
	LEFTARG = uncertain,
	RIGHTARG = real
);

-- Threshold operator function for >
CREATE OR REPLACE FUNCTION u_gt_thres(uncertain, u_pair) RETURNS boolean AS $$
DECLARE
    attr ALIAS FOR $1;
    value_threshold ALIAS FOR $2;
BEGIN
    RETURN (attr >% value_threshold.value) >= value_threshold.threshold;
END;
$$ LANGUAGE plpgsql;

-- Threshold operator for >
CREATE OPERATOR >? (
	PROCEDURE = u_gt_thres,
	LEFTARG = uncertain,
	RIGHTARG = u_pair
);


-------------------------------------------------------------------------------------------
-- (Uncertain < real) functions
-------------------------------------------------------------------------------------------
-- Probability operator function for <
CREATE OR REPLACE FUNCTION u_lt_prob(uncertain, real) RETURNS real AS $$
DECLARE
    attr ALIAS FOR $1;
    value ALIAS FOR $2;
    temp_arr real[];
BEGIN
	temp_arr[0] := value;
    RETURN u_prob(u_margin(attr, '{0}'), '{-inf}', temp_arr);
END;
$$ LANGUAGE plpgsql;

-- Probability operator for <
CREATE OPERATOR <% (
	PROCEDURE = u_lt_prob,
	LEFTARG = uncertain,
	RIGHTARG = real
);

-- Threshold operator function for <
CREATE OR REPLACE FUNCTION u_lt_thres(uncertain, u_pair) RETURNS boolean AS $$
DECLARE
    attr ALIAS FOR $1;
    value_threshold ALIAS FOR $2;
BEGIN
    RETURN (attr <% value_threshold.value) >= value_threshold.threshold;
END;
$$ LANGUAGE plpgsql;

-- Threshold operator for <
CREATE OPERATOR <? (
	PROCEDURE = u_lt_thres,
	LEFTARG = uncertain,
	RIGHTARG = u_pair
);

RESET search_path;

