--
-- pti_gist.sql : Installs Probability Threshold Indexing (PTI) into a DB.
--
-- Orion Project, Purdue University
-- http://orion.cs.purdue.edu/
--

SET search_path TO orion;

CREATE FUNCTION pti_consistent(internal, u_pair, int2)
RETURNS bool
AS ''
LANGUAGE internal;

CREATE FUNCTION pti_union(bytea, internal)
RETURNS uncertain
AS ''
LANGUAGE internal;

CREATE FUNCTION pti_compress(internal)
RETURNS internal
AS ''
LANGUAGE internal;

CREATE FUNCTION pti_decompress(internal)
RETURNS internal
AS ''
LANGUAGE internal;

CREATE FUNCTION pti_penalty(internal, internal, internal)
RETURNS internal
AS ''
LANGUAGE internal WITH (isstrict);

CREATE FUNCTION pti_picksplit(internal, internal)
RETURNS internal
AS ''
LANGUAGE internal;

CREATE FUNCTION pti_same(internal, internal, internal)
RETURNS internal
AS ''
LANGUAGE internal;

--
-- PTI Operator Class
--
-- Specifies the set of WHERE-clause operators that can be used with an index
-- (i.e., can be converted into an index-scan qualification). Also specifies
-- support procedures that are needed by the internal operations of GiST.
--
-- The operators associated with an operator class are identified by "strategy
-- numbers", which serve to identify the semantics of each operator within the
-- context of its operator class.  GiST indexes are completely flexible: they
-- do not have a fixed set of strategies at all. Instead, the "consistency"
-- support routine of each particular GiST operator class interprets the
-- strategy numbers however it likes.
--

CREATE OPERATOR CLASS pti_ops
  DEFAULT FOR TYPE uncertain USING gist AS

    -- uncertain, float comparisons
    OPERATOR   1  <?  (uncertain, u_pair)  ,
    OPERATOR   2  >?  (uncertain, u_pair)  ,

		-- TODO: Add more functions later

    -- GiST support procedures
    FUNCTION  1  pti_consistent (internal, u_pair, int2) ,
    FUNCTION  2  pti_union (bytea, internal) ,
    FUNCTION  3  pti_compress (internal) ,
    FUNCTION  4  pti_decompress (internal) ,
    FUNCTION  5  pti_penalty (internal, internal, internal) ,
    FUNCTION  6  pti_picksplit (internal, internal) ,
    FUNCTION  7  pti_same (internal, internal, internal) ;

RESET search_path;

