/*
 * dependency_set.h - Utility functions for rewriter module
 *
 * Orion Project, http://orion.cs.purdue.edu/
 */

#ifndef DEPENDENCY_SET_H_
#define DEPENDENCY_SET_H_

#include "postgres.h"
#include "nodes/makefuncs.h"

/* TODO(Sarvjeet): These values should be changed to something non-common
   (e.g. \001), once we rewrite the table output/printing functions */
const char *ATTR_SEPARATOR;
const char NOT_STORED_HERE_END_MARKER;

typedef struct DependencySet DependencySet;

struct DependencySet {
  char *attr_name;       /* The name of this attribute */
  int stored_in;         /* Where is this attribute stored (can be itself) */
  int dimension;         /* The dimension number (starts from 0) */
};

/* Populates the dependency set information for qry in output. The return
   value is the number of elements in output. output is allocated using palloc
   and it is the responsibility of the calling function to free the output */
int getDependencySet(const Query * const qry, DependencySet *output[]);

/* For debugging purposes */
void printDependencySet(DependencySet *d, int size);

#endif /*DEPENDENCY_SET_H_*/
