/*
 * oid_functions.c - Library functions to retrive oid of functions/types from
 * PostgreSQL backend. All functions cache their result, so multiple calls 
 * are cheap.
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "oid_functions.h"
#include "utils.h"

#define defineGetOidType(x) \
	Oid get##x##Oid() { \
		static Oid oid = 0; \
		if (oid == 0) \
			oid = getTypeOid(#x); \
		return oid; \
	}

#define defineGetOidFunc(x) \
	Oid get##x##Oid() { \
		static Oid oid = 0; \
		if (oid == 0) \
			oid = getFuncOid(#x); \
		return oid; \
	}

defineGetOidType(Uncertain)
defineGetOidFunc(U_product)
defineGetOidType(Int4)
defineGetOidType(Numeric)
defineGetOidFunc(U_margin)
defineGetOidFunc(U_floor)
defineGetOidFunc(U_floor_ex)
defineGetOidType(_int2)
defineGetOidType(_float4)
defineGetOidType(Float4)
defineGetOidFunc(U_real2hist)
defineGetOidFunc(_u_gt_uc)
defineGetOidFunc(_u_lt_uc)
defineGetOidFunc(_u_gte_uc)
defineGetOidFunc(_u_lte_uc)
defineGetOidFunc(_u_eq_uc)
defineGetOidFunc(_u_neq_uc)

defineGetOidFunc(_u_gt_uu)
defineGetOidFunc(_u_lt_uu)
defineGetOidFunc(_u_gte_uu)
defineGetOidFunc(_u_lte_uu)
defineGetOidFunc(_u_eq_uu)
defineGetOidFunc(_u_neq_uu)



defineGetOidType(Unknown)
defineGetOidType(CString)
defineGetOidFunc(Uncertain_parse)
defineGetOidType(UCon)
defineGetOidType(UDis)
defineGetOidType(UOrd)
defineGetOidType(UNom)
