/*
 * transform_insert.c - Rewrites the insert expression according to
 * Orion rewrite rules.
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "utils.h"
#include "nodes/makefuncs.h"
#include "utils/builtins.h"
#include "utils/snapmgr.h"
#include "utils/tqual.h"

#include "orion_bki.h"
#include "uncertain.h"
#include "oid_functions.h"
#include "debug.h"

/* Transform the Insert expr according to Orion rewriting rules */
Expr *transformOrionInsertExpr(Expr *expr, Oid final_type, Oid *initial_type, Oid relid, int attrno) {
//	DEBUG("inside transformOrionInsertExpr");
	int32 depid;
	int64 depdim;
	int16 dimnum;
	AttrNumber whereStored;
	
	FuncExpr *newFE;
    List *newArgs = NULL;
    Const *newC;
    char *typeString;
	bool has_error = false;
	
	if (ActiveSnapshot == NULL)
		ActiveSnapshot = GetTransactionSnapshot();
	else
		return expr;

	/* DEBUG */
//	orion_debug_pprint("Insert node before rewrite:", expr);

	if (isUncertainType(final_type)) {		
		if (isJointlyDistributed(relid,attrno, &dimnum, &depid, &depdim)) {
			whereStored = getAttrNum(depid);			
			if (whereStored == attrno) { /* Might need to rewrite this expr */
				if ((expr) && (IsA(expr, Const)) && 
				  (!(((Const*)expr)->constisnull)) && (*initial_type == getUnknownOid())) { 
					/* Make the type string */
					typeString = getTypeString(relid, depid);
					newC = makeConst(getCStringOid(), (int32) -1, -2,  
									DirectFunctionCall1(unknownin, CStringGetDatum(typeString)), 
									false, false);
					pfree(typeString);
					newArgs = lappend(newArgs, newC);
					/* Rewrite the insert string */
					((Const*)expr)->consttype = getCStringOid();
					newArgs = lappend(newArgs, expr);
					/* Make the func node */
					newFE = makeFuncExpr(getUncertain_parseOid(), 
										getUncertainOid(), 
										newArgs, 
										(CoercionForm) 0);
                    /* Make the relabel Node */
					expr = (Expr *) makeRelabelType((Expr *) newFE, final_type, (int32) -1, (CoercionForm) 2);
					*initial_type = final_type;
				}
			} else {
				if (!((expr) && (IsA(expr, Const)) && 
				  ((((Const*)expr)->constisnull))))
						has_error = true;
			}
		} else { /* Not jointly distributed, but still need to rewrite */
			if ((expr) && (IsA(expr, Const)) && 
				  (!(((Const*)expr)->constisnull)) && (*initial_type == getUnknownOid())) { 
					/* Make the type string */
					typeString = palloc(sizeof(char)*2);
					if (final_type == getUConOid()) typeString[0] = DIM_UCON;
					else if (final_type == getUDisOid()) typeString[0] = DIM_UDIS; 
					else if (final_type == getUNomOid()) typeString[0] = DIM_UNOM; 
					else typeString[0] = DIM_UORD; /* Not checking for any error case, but it should be one of these */
					typeString[1]='\0';
					
					newC = makeConst(getCStringOid(), (int32) -1, -2,  
									DirectFunctionCall1(unknownin, CStringGetDatum(typeString)), 
									false, false);
					pfree(typeString);
					newArgs = lappend(newArgs, newC);
					/* Rewrite the insert string */
					((Const*)expr)->consttype = getCStringOid();
					newArgs = lappend(newArgs, expr);
					/* Make the func node */
					newFE = makeFuncExpr(getUncertain_parseOid(), 
                                         getUncertainOid(), 
                                         newArgs, 
                                         (CoercionForm) 0);
                    /* Make the relabel Node */
					expr = (Expr *) makeRelabelType((Expr *) newFE, final_type, (int32) -1, (CoercionForm) 2);
					*initial_type = final_type;
				}
		}
	}
//	DEBUG("done");
	/* DEBUG */
//	orion_debug_pprint("Insert node after rewrite:", expr);
	
	ActiveSnapshot = NULL;
	if(has_error)
		elog(ERROR, "Orion: Trying to insert an uncertain value into an attribute that is stored elsewhere (as a joint distribution)");
	return expr;	
}
