#include "postgres.h"
#include "nodes/makefuncs.h"
#include "nodes/print.h"
#include "catalog/namespace.h"
#include "catalog/pg_type.h"
#include "executor/spi.h"
#include "utils/fmgroids.h"
#include "utils/numeric.h"
#include "utils/snapmgr.h"

//#include "stdlib.h"
#include "uncertain.h"
#include "orion_bki.h"
#include "oid_functions.h"
#include "utils.h"
#include "errors.h"
#include "debug.h"

/*
TODO :
0) make certain uncertain in case of uc > cc : DONE
1) support uu <> uu : DONE
2) change target entry resname as follows : depending upon marg, it should target entry number _ target entry number : DONE
*/

/************** Begin structures *********************/

/* Structure for storing Attribute Map */
typedef struct AttributeMap AttributeMap;
int countMap = 0; // for debugging
struct AttributeMap {
	int index; //just for debugging
	Index varno; /* index of this var's relation in the range table */
	Oid resorigtbl; /* Oid of original table */
		
	AttrNumber varattno; /* attribute number of this var in the table*/
	Oid vartype; /* Type of this variable */
	
	int32 depid; /* The depid of this attr, -1 if its not in a depid */
	int origdimnum; /*dimension number in the original pdf, starting from 1 */
	
	Expr *expr; /*stores value to which this attribute finally boils down to. If NULL that means that the original targetEntry->expr remains the same*/
	int32 maxdim; /*stores the dimension of pdf stored on this attribute.*/
	char marg_arg[100]; /*stores which dimensions to keep from the final pdf, starting from 0*/
	char resname[100]; /*stores the new resname for the target entry*/

	TargetEntry *node; /* points to the TargetEntry which requires this attribute */
	int dimension; /* the dimension number of this attribute (in the final/resulting distribution, starting from 1 */

	AttributeMap *mappedTo; /*the attribute which stores the pdf belonging to this attribute */
	AttributeMap *next; /* Pointer to next node */
};
AttributeMap *attrMap = NULL;
AttributeMap *endAttrMap = NULL;
AttributeMap *finalAttrMap = NULL; //stores the attribute to which the where attributes are merged

AttributeMap *depidSeen(int32 depid, int32 varno);
AttributeMap *attrSeen(Oid resorigtbl, Index varno, AttrNumber varattno, AttributeMap *attr);
AttributeMap *attrSeenSelfJoin(Oid resorigtbl, Index varno, AttrNumber varattno, AttributeMap *attr);
void addDim(AttributeMap *attr, int depdim);
void mergeAttributeMap(AttributeMap *firstAttr, AttributeMap *secondAttr);
AttributeMap *updateWhereAttrMap(Query *qry, Node *node, AttributeMap *finalAttr);
Node* applyFloors(Query *qry, Node *node);
void createAttrMap(Query *qry);
void freeAttrMap(void);
void processJointDependencies(void);
Query* updateTEs(Query *subQry);
Query *transformOrionSelectStmt(Query *qry);
void getFloorExpr(Query *qry, Node *node, AttributeMap **finalAttr, char *floor_arg, char *floor_type);
AttributeMap *addWhereAttrMap(Query *qry, Var *currVar, AttributeMap *localFinalAttr);
void printAttrMap(void);
Expr *makeUncertain(Var *currV);
Query *returnNonzeroTuples(Query *subQry, BoolExpr *newAND);
bool hasUncertain(List *rtablelist);

/*
returns an attributeMap if depid has been seen before
the returned attributeMap is the attribute which keeps the pdf
*/
AttributeMap *depidSeen(int32 depid, int32 varno) {
	AttributeMap *temp;
	
	for(temp = attrMap; temp != NULL; temp = temp->next) {
		if (temp->depid == depid) { /* Found depid in mlist */
			if(varno == -1)
				return temp->mappedTo;
			else if(varno == temp->mappedTo->varno)
				return temp->mappedTo;
		}
	}
	return NULL;	
}

/*
returns an attributeMap if attr has been seen before
the returned attributeMap is the attribute which has the same table/column value
if attr is given, only attributes that came before attr will be searched.
if attr is null, all attributes will be searched
	
In the case of self join, each table is considered different.
*/
AttributeMap *attrSeen(Oid resorigtbl, Index varno, AttrNumber varattno, AttributeMap *attr) {
//	DEBUG("attrSeen");
	AttributeMap *temp;

	for(temp = attrMap; (temp != NULL)&&(temp != attr) ; temp = temp->next) {
		if ((temp->varno == varno) && (temp->varattno == varattno)){ /* Found attribute */
		//if ((temp->resorigtbl == resorigtbl) && (temp->varattno == varattno)){ /* Found attribute */
			DEBUG("attribute was seen");
			DEBUG("attrSeen done");
			return temp;
		}
	}
	DEBUG("attr not seen");
	DEBUG("attrSeen done");
	return NULL;
}

AttributeMap *attrSeenSelfJoin(Oid resorigtbl, Index varno, AttrNumber varattno, AttributeMap *attr) {
//	DEBUG("attrSeen");
	AttributeMap *temp;

	for(temp = attrMap; (temp != NULL)&&(temp != attr) ; temp = temp->next) {
		//if ((temp->varno == varno) && (temp->varattno == varattno)){ /* Found attribute */
		if ((temp->resorigtbl == resorigtbl) && (temp->varno != varno) && (temp->varattno == varattno)){ /* Found attribute */
			DEBUG("attribute was seen");
			DEBUG("attrSeen done");
			return temp;
		}
	}
	DEBUG("attr not seen");
	DEBUG("attrSeen done");
	return NULL;
}

void addDim(AttributeMap *attr, int depdim) {
	char *ans, *token;
	bool first = true;
	char temp[10];
	
	ans = palloc(200);
	ans[0]='\0';
	
	token = strtok (attr->marg_arg," ,");
	while (token != NULL) {
		if (!first)
			strcat(ans,",");
		else
			first = false;

		sprintf(temp, "%d" , atoi(token)+depdim);
		strcat(ans, temp);
    	token = strtok (NULL, " ,");
	}
	snprintf(attr->marg_arg, 100, "%s", ans);
	pfree(ans);
}

//it is assumed that firstAttr and secondAttr both map to itself.
//it is also assumed that both attr's node are not null. Atleast one is nonnull
void mergeAttributeMap(AttributeMap *firstAttr, AttributeMap *secondAttr) {
//	DEBUG("mergeAttributeMap");
	bool oneIsNull = false;
	FuncExpr *newFExpr = NULL;
	List *newArgs = NULL;
	Expr * expr1 = NULL;
	Expr * expr2 = NULL;
	AttributeMap *temp;
	if(firstAttr->node == NULL) {
		AttributeMap *temp = firstAttr;
		firstAttr = secondAttr;
		secondAttr = temp;
	}
	if(secondAttr->node == NULL)
		oneIsNull = true;

	
//	DEBUG("merging %d and %d", firstAttr->index, secondAttr->index);
//	if(firstAttr->mappedTo != firstAttr)
//		DEBUG("mergeAttributeMap got wrong firstAttr. It doesnt map to itself");
//	if(secondAttr->mappedTo != secondAttr)
//		DEBUG("mergeAttributeMap got wrong secondAttr. It doesnt map to itself");

	//find dimension of the pdf at firstAttr, add that to marg_arg of secondAttr

	addDim(secondAttr, firstAttr->maxdim);
	//update dimension of all the affect attributes. Do we need it? Yes we do
	for(temp = attrMap; temp != NULL; temp = temp->next) {
		if (temp->mappedTo == secondAttr) {
			temp->mappedTo = firstAttr;
			temp->dimension += firstAttr->maxdim;
		}
	}
	firstAttr->maxdim += secondAttr->maxdim;
	//product of pdf at firstAttr and secondAttr
	
	if(firstAttr->expr == NULL)
		expr1 = copyObject(firstAttr->node->expr);
	else
		expr1 = firstAttr->expr;

	newArgs = lappend(newArgs, expr1);

	if(secondAttr->expr == NULL)
		expr2 = copyObject(secondAttr->node->expr);
	else
		expr2 = secondAttr->expr;

	newArgs = lappend(newArgs, expr2);

	newFExpr = makeFuncExpr(getU_productOid(), 
							getUncertainOid(), 
							newArgs, 
							(CoercionForm) 0);

	firstAttr->expr = (Expr *)newFExpr;
	if(!oneIsNull) {
		strcat(firstAttr->marg_arg, ",");
		strcat(firstAttr->marg_arg, secondAttr->marg_arg);
		strcat(firstAttr->resname, "_");
		strcat(firstAttr->resname, secondAttr->resname);
	}
	secondAttr->expr = (Expr *) makeConst(getUncertainOid(), (int32) -1, -1, (Datum) 0, true, false);
	secondAttr->mappedTo = firstAttr;
	//printAttrMap();
//	DEBUG("mergeAttributeMap finished");
}


/* Make a Certain/Precise Variable Uncertain - Returns NULL if unsuccessful*/
Expr *makeUncertain(Var *currV) {
//	DEBUG("makeUncertain");
	Oid float4Oid = getFloat4Oid();
	Oid argType;
	Oid funcOid = InvalidOid;
	List *newArgs;
//	Const *newC;
	FuncExpr *newFE;
	
	
	argType = currV->vartype;

	/* Find the appropriate function to convert it to float4 */
	if (argType != float4Oid)
		funcOid = getConvertFloat4Oid(argType);
	
	if (funcOid == InvalidOid) {
		char *errorHint = palloc(100);
		snprintf(errorHint, 100, "makeUncertain - Failed while trying to convert Oid type (%d) to float", argType);
		err_custom("Orion query rewriter failed", errorHint);
		pfree(errorHint);
		return NULL;
	}

	
	/* Convert the var --> u_margin(u_real2hist(convertfunc(var)), '{0}') */
	newArgs = lappend(NULL, currV);
	if (argType != float4Oid) {
		newFE = makeFuncExpr(funcOid, float4Oid, newArgs, (CoercionForm) 2);
		newArgs = lappend(NULL, newFE);
	}

	newFE = makeFuncExpr(getU_real2histOid(), getUncertainOid(), newArgs, (CoercionForm) 0);
//	newArgs = lappend(NULL, newFE);	
//	newC = makeConst(get_int2Oid(), (int32) -1, -1, (Datum) 0, false, false);
//	newArgs = lappend(newArgs, newC);
//	newFE = makeFuncExpr(getU_marginOid(), getUncertainOid(), newArgs, (CoercionForm) 0);

	return (Expr *)newFE;
}



AttributeMap *addWhereAttrMap(Query *qry, Var *currVar, AttributeMap *localFinalAttr) {
//	DEBUG("addWhereAttrMap");
	RangeTblEntry *nodeTable;
	int16 dimnum = -1;
	int32 depid = -1;
	int64 depdim = -1;

	AttributeMap *attr = NULL;


	nodeTable = (RangeTblEntry *) list_nth(qry->rtable, currVar->varno - 1);
	attr = attrSeen(nodeTable->relid, currVar->varno, currVar->varattno, NULL);
//	if(attr!=NULL) { // attr is in the targetlist, do nothing
//		DEBUG("attr %s seen", attr->node->resname);
//	}

	if(attr == NULL) {
		//attr not seen before and depid not seen before
//		DEBUG("variable not seen");
		attr = endAttrMap;
		attr->next = palloc(sizeof(AttributeMap));
		attr = attr->next;
		endAttrMap = attr;
		attr->index = countMap++;

		attr->varno = currVar->varno;
		attr->resorigtbl = nodeTable->relid;

		attr->varattno = currVar->varattno;
		attr->vartype = currVar->vartype;
		attr->depid = -1;
		attr->node = NULL;
		attr->dimension = 1; 
		attr->maxdim = 1; 
		attr->origdimnum = 1;
		attr->next = NULL;
		attr->expr = NULL;
		attr->mappedTo = attr;
		snprintf(attr->marg_arg, 100, "0");

		if (isJointlyDistributed(nodeTable->relid, currVar->varattno, &dimnum, &depid, &depdim)) {
			//check if this depid has been seen before
			AttributeMap* attr2 = depidSeen(depid, -1);
			if(attr2 != NULL) { //depid seen before, do nothing
				attr->dimension = dimnum + (attr2->dimension - attr2->origdimnum); //dimension in actual pdf + dimensions because of u_product
				attr->origdimnum = dimnum;
				attr->mappedTo = attr2->mappedTo;
				attr->maxdim = attr2->mappedTo->maxdim;
			}
			else{//jointly distributed but not seen before
				AttrNumber jointAttrNum = getAttrNum(depid);
				attr->expr = (Expr *) makeVar(	attr->varno, 
												jointAttrNum, 
												getVarType(attr->resorigtbl, jointAttrNum),  
												(int32) -1, 
												(Index) 0);
				snprintf(attr->marg_arg, 100, "%d", (dimnum-1));
				attr->maxdim = depdim;
				attr->dimension = dimnum;
				//This node has to be merged with some targetentry as right now it's pdf is not in any targetentry
				attr->depid = depid;
				attr->origdimnum = dimnum;
			}
		}
		else {
			if(isUncertainType(currVar->vartype)) {
				attr->expr = (Expr *) makeVar(	attr->varno, 
												attr->varattno, 
												getVarType(attr->resorigtbl, attr->varattno),  
												(int32) -1, 
												(Index) 0);
			}
			else {
				//make it uncertain
				attr->expr = makeUncertain(currVar);
			}
		}
	}
	if(localFinalAttr == NULL) //that means its the first attribute in the OR clause
		localFinalAttr = attr->mappedTo;
	if(localFinalAttr != attr->mappedTo) //merge them
		mergeAttributeMap(localFinalAttr, attr->mappedTo);

//	DEBUG("addWhereAttrMap finished");

	return localFinalAttr;

}



AttributeMap *updateWhereAttrMap(Query *qry, Node *node, AttributeMap *localFinalAttr) {
//	DEBUG("updateWhereAttrMap");
//	Node *node = qry->jointree->quals;
	OpExpr *oExpr;
	BoolExpr *bExpr;
	Oid u_gt_uc = get_u_gt_ucOid();
	Oid u_lt_uc = get_u_lt_ucOid();
	Oid u_gt_uu = get_u_gt_uuOid();
	Oid u_lt_uu = get_u_lt_uuOid();
	Oid u_gte_uc = get_u_gte_ucOid();
	Oid u_lte_uc = get_u_lte_ucOid();
	Oid u_gte_uu = get_u_gte_uuOid();
	Oid u_lte_uu = get_u_lte_uuOid();
	
	Oid u_eq_uc = get_u_eq_ucOid();
	Oid u_neq_uc = get_u_neq_ucOid();
	Oid u_eq_uu = get_u_eq_uuOid();
	Oid u_neq_uu = get_u_neq_uuOid();

	
	void *temp;
	Var *currVar;
	Node *currNode;
	ListCell *l;

	if (IsA(node,OpExpr)) { /* Operator */
//		DEBUG("Operator");
		oExpr = (OpExpr *)node;
		if ((oExpr->opfuncid == u_gt_uc) || 
			(oExpr->opfuncid == u_gte_uc) || 
			(oExpr->opfuncid == u_lte_uc) || 
			(oExpr->opfuncid == u_eq_uc) || 
			(oExpr->opfuncid == u_neq_uc) || 
			(oExpr->opfuncid == u_lt_uc)||  /* uncertain <> float4 */
			(oExpr->opfuncid == u_gt_uu) || 
			(oExpr->opfuncid == u_gte_uu) || 
			(oExpr->opfuncid == u_lte_uu) || 
			(oExpr->opfuncid == u_eq_uu) || 
			(oExpr->opfuncid == u_neq_uu) || 
			(oExpr->opfuncid == u_lt_uu)) { /* uncertain >< uncertain */

//			DEBUG("uncertain <>= something");
			temp = list_nth(oExpr->args, 0);
			currVar = (Var *) ((RelabelType*)temp)->arg; //Assumed that temp is RelabelType and arg is Var type
			localFinalAttr = addWhereAttrMap(qry, currVar, localFinalAttr);

			temp = list_nth(oExpr->args, 1);
			if(IsA(temp, FuncExpr)) {
				if(!IsA(list_nth(((FuncExpr*)temp)->args, 0), Const)) {
//					DEBUG("Second variable not a constant");
					currVar = (Var *)list_nth(((FuncExpr*)temp)->args, 0);
					localFinalAttr = addWhereAttrMap(qry, currVar, localFinalAttr);
				}
			}
			else {
//				DEBUG("Second variable not a constant");
				//Second Arguement is a variable too
				currVar = (Var *) ((RelabelType*)temp)->arg; //Assumed that temp is RelabelType and arg is Var type
				//printf(currVar->varno);
				localFinalAttr = addWhereAttrMap(qry, currVar, localFinalAttr);
			}
//			DEBUG("updateWhereAttrMap finished");
			return localFinalAttr;
		}
	} 
	else if (IsA(node, BoolExpr)) { /* Boolean expression */
//		DEBUG("Boolean Expression");
		bExpr = (BoolExpr *) node;
		/*if AND, call the same function for all the arguements. 
			The first arguement will return the attribute to which the atributes were merged if they were merged at all, 
			otherwise it will return the attribute on which this part (as in the first disjunction) of floor will be applied.
			After that, each call will be sent this first attribute so that in case of no final attribute this attribute will be used
		*/
		if (bExpr->boolop == AND_EXPR) {// doesnt return anything since the AND arguments do not make attributes depended upon each other
//			DEBUG("AND_EXPR");
			//TODO: Check if the whole and expression has any uncertainity. If yes, then merge attributes, otherwise not.
			//TODO: If it has, after merging, remove it from Quals
			foreach(l,bExpr->args) {
				AttributeMap *final = NULL;
				currNode = (Node *)lfirst(l);
				final = updateWhereAttrMap(qry, currNode, NULL);
				if((final != NULL) && (final->node == NULL)) { //merge it with first attributeMap
					mergeAttributeMap(attrMap->mappedTo, final);
					//TODO: remove it from quals
				}
			}
		}
		else if (bExpr->boolop == OR_EXPR) {
//			DEBUG("OR_EXPR");
			AttributeMap *final = NULL;
			foreach(l,bExpr->args) {
				currNode = (Node *)lfirst(l);
				final = updateWhereAttrMap(qry, currNode, final);
			}
			//if the OR attributes were not merged to some TE attributes, then merge it with localFinalAttr
			if ((final != NULL) && (final->node == NULL)) {
				if(localFinalAttr != NULL) {
					mergeAttributeMap(localFinalAttr, final);
//					DEBUG("updateWhereAttrMap finished");
					return localFinalAttr;
				}
				//else return final
			}
//			DEBUG("updateWhereAttrMap finished");
			return final;
		}
	}
//	DEBUG("updateWhereAttrMap finished");
	return NULL;
}

/*
	Converts the query tree quals in a CNF form.
	Fix this function so when its already in CNF form, it doesnt change anything.
*/
static Node * convertToCNF(Node * node) {
	OpExpr *oExpr;
	BoolExpr *bExpr;
	ListCell *l;
	ListCell *l2;

	if (IsA(node, BoolExpr)) { /* Boolean expression */
		//printf("Got a Boolean Expression\n");

		bExpr = (BoolExpr *)node;
		if (bExpr->boolop == AND_EXPR) { /* divide and conqueror */
			BoolExpr *newAND = makeNode(BoolExpr);
			bool compressed = false;
			//printf("AND\n");
			newAND->boolop = AND_EXPR;

			foreach(l, bExpr->args) {
				BoolExpr *newBExpr = (BoolExpr*) convertToCNF((Node *)lfirst(l));
				newAND->args = lappend(newAND->args, newBExpr);
			}
			
			/* compress the tree for all ANDs */
			while (! compressed) {
				ListCell * prev = NULL;
				compressed = true;
				foreach(l, newAND->args) {
					BoolExpr *l_node = (BoolExpr *)lfirst(l);
					if(l_node->boolop == AND_EXPR) {
						compressed = false;
						newAND->args = list_delete_cell(newAND->args, l, prev);
						foreach(l2, l_node->args) {
							newAND->args = lappend(newAND->args, (Node *)lfirst(l2));
						}
					}
					prev = l;
				}
			}
			return (Node*) newAND;
		}
		if (bExpr->boolop == OR_EXPR) {
			bool compressed = false;
			ListCell *prev = NULL;

			//printf("OR\n");
			/* first compress the tree for all ORs */
			while (! compressed) {
				ListCell * prev = NULL;
	
				compressed = true;
				foreach(l, bExpr->args) {
					BoolExpr *l_node = (BoolExpr *)lfirst(l);
					if(l_node->boolop == OR_EXPR) {
						compressed = false;
						bExpr->args = list_delete_cell(bExpr->args, l, prev);
						foreach(l2, l_node->args) {
							bExpr->args = lappend(bExpr->args, (Node *)lfirst(l2));
						}
					}
					prev = l;
				}
			}
			/* pick an AND child and then convert */
			foreach(l, bExpr->args) {
				BoolExpr *l_node = (BoolExpr *)lfirst(l);
				if(l_node->boolop == AND_EXPR) {
					BoolExpr *newAND = makeNode(BoolExpr);
					bExpr->args = list_delete_cell(bExpr->args, l, prev);
					newAND->boolop = AND_EXPR;
				
					foreach(l2, l_node->args) {
						BoolExpr *newOR = makeNode(BoolExpr);
						newOR->boolop = OR_EXPR;
						newOR->args = lappend(newOR->args, copyObject(node));
						newOR->args = lappend(newOR->args, (Node *)lfirst(l2));
						newAND->args = lappend(newAND->args, newOR);
					}
					return convertToCNF((Node*) newAND);
				}
				prev = l;
			}
		}
	}
	if (IsA(node,OpExpr)) { /* Operator */
		oExpr = (OpExpr *)node;	
		//printf("%d\n", oExpr->opfuncid);
	}
	return node;
}


void printAttrMap() {
	AttributeMap * temp = NULL;
	AttributeMap * temp2 = NULL;
	int count = 1;
	int count2 = 1;

	DEBUG("ATTRIBUTE MAP");
	for(temp = attrMap; temp != NULL; temp = temp->next) {
		DEBUG("ATTRIBUTE %d : ", count++);
		DEBUG("\tdimension : %d", temp->dimension);
		DEBUG("\tindex : %d", temp->index);
		
		if(temp->node != NULL)
			DEBUG("\t resname : %s", temp->node->resname);
		DEBUG("\t marg_arg : %s", temp->marg_arg);
		for(temp2 = attrMap; temp2 != NULL; temp2 = temp2->next) {
			if(temp2==temp->mappedTo) {
				DEBUG("\t points to : %d", count2);
				break;
			}
			count2++;
		}
		if(temp2==temp->mappedTo)
			DEBUG("\tmaxdim : %d", temp2->maxdim);
		count2 = 1;
	}
}

void getFloorExpr(Query *qry, Node *node, AttributeMap **finalAttr, char *floor_arg, char *floor_type) {
//	DEBUG("getFloorExpr");
	OpExpr *oExpr;
	BoolExpr *bExpr;
	Oid u_gt_uc = get_u_gt_ucOid();
	Oid u_lt_uc = get_u_lt_ucOid();
	Oid u_gte_uc = get_u_gte_ucOid();
	Oid u_lte_uc = get_u_lte_ucOid();
	Oid u_eq_uc = get_u_eq_ucOid();
	Oid u_neq_uc = get_u_neq_ucOid();

	Oid u_gt_uu = get_u_gt_uuOid();
	Oid u_lt_uu = get_u_lt_uuOid();
	Oid u_gte_uu = get_u_gte_uuOid();
	Oid u_lte_uu = get_u_lte_uuOid();
	Oid u_eq_uu = get_u_eq_uuOid();
	Oid u_neq_uu = get_u_neq_uuOid();

	Var *currVar = NULL;
	RangeTblEntry *nodeTable;
	ListCell *l;
	void *temp;
	Const *cExpr;
	AttributeMap *attr = NULL;
	AttributeMap *attr2 = NULL;
	char *const_value;
	int i;
	int dimension = 0;
	int dimension2 = -1;
	bool certainVar = false;
	floor_arg[0] = '\0';
	floor_type[0] = '\0';

	if (IsA(node,OpExpr)) { /* Operator */
		oExpr = (OpExpr *)node;	
		if ((oExpr->opfuncid == u_gt_uc) || 
			(oExpr->opfuncid == u_gte_uc) || 
			(oExpr->opfuncid == u_lte_uc) || 
			(oExpr->opfuncid == u_eq_uc) || 
			(oExpr->opfuncid == u_neq_uc) || 
			(oExpr->opfuncid == u_lt_uc) || /* uncertain <>= float4 */
			(oExpr->opfuncid == u_gt_uu) || 
			(oExpr->opfuncid == u_gte_uu) || 
			(oExpr->opfuncid == u_lte_uu) || 
			(oExpr->opfuncid == u_eq_uu) || 
			(oExpr->opfuncid == u_neq_uu) || 
			(oExpr->opfuncid == u_lt_uu)) { /* uncertain >< uncertain */



			//set floor_type first
			if((oExpr->opfuncid == u_neq_uc) || (oExpr->opfuncid == u_neq_uu))
				strcat(floor_type, "0");
			else if((oExpr->opfuncid == u_eq_uc) || (oExpr->opfuncid == u_eq_uu))
				strcat(floor_type, "1");
			else if((oExpr->opfuncid == u_lt_uc) || (oExpr->opfuncid == u_lt_uu))
				strcat(floor_type, "2");
			else if((oExpr->opfuncid == u_lte_uc) || (oExpr->opfuncid == u_lte_uu))
				strcat(floor_type, "3");
			else if ((oExpr->opfuncid == u_gt_uc) || (oExpr->opfuncid == u_gt_uu))
				strcat(floor_type, "4");
			else if((oExpr->opfuncid == u_gte_uc) || (oExpr->opfuncid == u_gte_uu))
				strcat(floor_type, "5");

			temp = list_nth(oExpr->args, 0);
			currVar = (Var *) ((RelabelType*)temp)->arg; //Assumed that temp is RelabelType and arg is Var type
			nodeTable = (RangeTblEntry *) list_nth(qry->rtable, currVar->varno - 1);
			attr = attrSeen(nodeTable->relid, currVar->varno, currVar->varattno, NULL);
			if (attr != NULL){
				*finalAttr = attr->mappedTo;
				dimension = attr->dimension;
			}
			else {
				int16 dimnum = -1;
				int32 depid = -1;
				int64 depdim = -1;
				if (isJointlyDistributed(nodeTable->relid, currVar->varattno, &dimnum, &depid, &depdim)) {
					//check if this depid has been seen before
					attr = depidSeen(depid, -1);
					*finalAttr = attr->mappedTo;
					dimension = dimnum + (attr->dimension - attr->origdimnum);
				}
			}
			
		}
//		DEBUG("dimension = %d", dimension);
		if ((oExpr->opfuncid == u_gt_uc) ||
			(oExpr->opfuncid == u_gte_uc) || 
			(oExpr->opfuncid == u_lte_uc) || 
			(oExpr->opfuncid == u_eq_uc) || 
			(oExpr->opfuncid == u_neq_uc) || 
			(oExpr->opfuncid == u_lt_uc)) { /* uncertain <> float4 */

			temp = list_nth(oExpr->args,1);
			if (!IsA(temp, FuncExpr)) {
				return;
			}


			cExpr = (Const *) list_nth(((FuncExpr *)temp)->args,0);
			if (!IsA(cExpr, Const)) {
				//it was a certain variable
				certainVar = true;
				//return;
//				DEBUG("certainVar");
			}
			else {
				const_value = getConstValue(cExpr);
				strcat(floor_arg, "{");
				for(i = 0; i < attr->mappedTo->maxdim;  i++)
					strcat(floor_arg, (((i+1) != dimension)? "0,":"1," ));

				strcat(floor_arg, const_value);

				pfree(const_value);
				strcat(floor_arg,"}");
			}
		} 
		if ((oExpr->opfuncid == u_gt_uu) || 
			(oExpr->opfuncid == u_gte_uu) || 
			(oExpr->opfuncid == u_lte_uu) || 
			(oExpr->opfuncid == u_eq_uu) || 
			(oExpr->opfuncid == u_neq_uu) || 
			(oExpr->opfuncid == u_lt_uu) || /* uncertain >< uncertain */
			(certainVar)) { 

			temp = list_nth(oExpr->args, 1);
			if(certainVar)
				currVar = (Var *)list_nth(((FuncExpr*)temp)->args, 0);
			else
				currVar = (Var *) ((RelabelType*)temp)->arg; //Assumed that temp is RelabelType and arg is Var type
			nodeTable = (RangeTblEntry *) list_nth(qry->rtable, currVar->varno - 1);
			attr2 = attrSeen(nodeTable->relid, currVar->varno, currVar->varattno, NULL);
			if (attr2 != NULL){
				dimension2 = attr2->dimension;
			}
			else {
//				DEBUG("second variable not seen");
				int16 dimnum = -1;
				int32 depid = -1;
				int64 depdim = -1;
				if (isJointlyDistributed(nodeTable->relid, currVar->varattno, &dimnum, &depid, &depdim)) {
					//check if this depid has been seen before
//					DEBUG("second variable jointly distributed");
					attr2 = depidSeen(depid, -1);
					dimension2 = dimnum + (attr2->dimension - attr2->origdimnum);
//					DEBUG("dimension2 = %d", dimension2);
				}
			}
			strcat(floor_arg, "{");
			for(i = 0; i < attr->mappedTo->maxdim;  i++) {
				if((i+1) == dimension) 
					strcat(floor_arg, "1,");
				else if((i+1) == dimension2) 
					strcat(floor_arg, "-1,");
				else
					strcat(floor_arg, "0,");
			}
			strcat(floor_arg, "0");
			strcat(floor_arg,"}");
		}
	} 
	else if (IsA(node, BoolExpr)) { // Boolean expression. It can not be AND 
		bExpr = (BoolExpr *)node;	

		if (bExpr->boolop == OR_EXPR) {
			int i = 0;
			foreach(l,bExpr->args) {
				Node *currNode = (Node *)lfirst(l);
				char* temp_str = palloc(sizeof(char) * 200);
				char* temp_str_type = palloc(sizeof(char) * 200);
				char * temp_str2 = palloc(sizeof(char) * 200);
				char * temp_str2_type = palloc(sizeof(char) * 200);
				strcpy(temp_str, floor_arg);
				strcpy(temp_str_type, floor_type);
				pfree(floor_arg);
				pfree(floor_type);
				floor_arg = palloc(sizeof(char) * 200); /* TODO: Don't hard code this */
				floor_type = palloc(sizeof(char) * 200); /* TODO: Don't hard code this */
				floor_arg[0] = '\0';
				floor_type[0] = '\0';
				temp_str2[0] = '\0';
				temp_str2_type[0] = '\0';
				getFloorExpr(qry, currNode, finalAttr, temp_str2, temp_str2_type);
				if(i == 0) {
					sprintf(floor_arg, "%s", temp_str2);
					sprintf(floor_type, "%s", temp_str2_type);
					i++;
				}
				else {
					sprintf(floor_arg, "%s%s%s", temp_str, ",", temp_str2);
					sprintf(floor_type, "%s%s%s", temp_str_type, ",", temp_str2_type);
				}
				pfree(temp_str2);
				pfree(temp_str2_type);
				pfree(temp_str);
				pfree(temp_str_type);
			}
		}
	}
//	DEBUG("floor is %s", floor_arg);
//	DEBUG("getFloorExpr finished");

}

Node* applyFloors(Query *qry, Node *node) {
//	DEBUG("applyFloors");
	Node* residueNode = node; //Null if the floor was applied successful
	BoolExpr *bExpr;
	char *floor_arg = palloc(sizeof(char) * 200);
	char *floor_type = palloc(sizeof(char) * 200);
//	Node *node = qry->jointree->quals;
	AttributeMap *finalAttr = NULL;
	List *newArgs = NULL;
	Expr * expr1 = NULL;
	Const *newC;
	Const *newC_type;
	char *temp_str;
	char *temp_str_type;
	
	if (IsA(node,OpExpr)) { /* Operator */
		getFloorExpr(qry, node, &finalAttr, floor_arg, floor_type);
	}
	else if (IsA(node, BoolExpr)) { /* Boolean expression. It can not be AND */
		bExpr = (BoolExpr *)node;	
		if (bExpr->boolop == OR_EXPR) {
			getFloorExpr(qry, node, &finalAttr, floor_arg, floor_type);
		}
		else if (bExpr->boolop == AND_EXPR) {
			ListCell   *l;
//			List * newArgs = NULL;
			foreach(l, bExpr->args) {
				Node *currNode = (Node *)lfirst(l);
				if(applyFloors(qry, currNode))
					lappend(newArgs, currNode);
			}
			if(newArgs != NULL) {
				bExpr->args = newArgs;
			}
				
			pfree(floor_arg);
			return residueNode;
		}
	}
	if(finalAttr != NULL) {
		residueNode = NULL;
		temp_str = palloc(strlen(floor_arg) + 3);
		temp_str_type = palloc(strlen(floor_type) + 3);
		if(finalAttr->expr == NULL)
			expr1 = copyObject(finalAttr->node->expr);
		else
			expr1 = finalAttr->expr;
		newArgs = lappend(newArgs, expr1);
		sprintf(temp_str, "{%s}", floor_arg);
		sprintf(temp_str_type, "{%s}", floor_type);

//		DEBUG("Floor : %s", floor_arg);
//		DEBUG("Floor Types: %s", floor_type);

		newC_type = makeConst(get_int2Oid(), (int32) -1, -1,
							OidFunctionCall3(F_ARRAY_IN, CStringGetDatum(temp_str_type),
							ObjectIdGetDatum(INT2OID), -1),
							false, false);
		newArgs = lappend(newArgs, newC_type);

		newC = makeConst(get_float4Oid(), (int32) -1, -1,
							OidFunctionCall3(F_ARRAY_IN, CStringGetDatum(temp_str),
							ObjectIdGetDatum(FLOAT4OID), -1),
							false, false);

		newArgs = lappend(newArgs, newC);

		finalAttr->expr = (Expr *)makeFuncExpr(getU_floor_exOid(),
										getUncertainOid(),
										newArgs,
										(CoercionForm) 0);
	}

	pfree(floor_arg);
	return residueNode;
//	DEBUG("applyFloors finished");
}


void createAttrMap(Query *qry) {
//	DEBUG("createAttrMap");
	ListCell   *l;
	TargetEntry *attr;
	AttributeMap *currAttr = NULL;
	Var *currVar;
	RangeTblEntry *nodeTable = NULL;
	int16 dimnum;
	int32 depid;
	int64 depdim;
	AttributeMap *mappedTo = NULL;
	int count=0;
	int index;
	//From each target entry, pick the uncertain variables and create a corresponding AttributeMap
	foreach(l, qry->targetList) {
		count++;
		attr = ((TargetEntry *) lfirst(l));

		if ((IsA(attr->expr,Var) &&  								// Is this a variable  // if its an expression, go through each variable in that expression and see if there is any uncertain variable
			(isUncertainType(((Var *)attr->expr)->vartype)))) { 	// Is this attr Uncertain ?

			currVar = (Var *)(attr->expr);

			if (attrMap == NULL)
				currAttr = attrMap = palloc(sizeof(AttributeMap)); /* palloc will not work as SPI_finish restores the stack (?) */
			else {
				currAttr->next = palloc(sizeof(AttributeMap));
				currAttr = currAttr->next;
			}
			//char temp_str[100];
			sprintf(currAttr->resname, "%d", count);
			//strcat(currAttr->resname, temp_str);
			currAttr->index = countMap++;

			nodeTable = (RangeTblEntry *) list_nth(qry->rtable, currVar->varno - 1);

			currAttr->varno = currVar->varno;
			currAttr->resorigtbl = nodeTable->relid;

			currAttr->varattno = currVar->varattno;
			currAttr->vartype = currVar->vartype;
			currAttr->depid = -1;
			currAttr->node = attr;
			currAttr->dimension = 1; 
			currAttr->maxdim = 1; 
			currAttr->origdimnum = 1;
			currAttr->next = NULL;
			currAttr->expr = NULL;
			currAttr->mappedTo = currAttr;
			snprintf(currAttr->marg_arg, 100, "0");


			//if the attribute was seen before, dont need to check if its jointly distributed since the previous attribute took care of it
			mappedTo = attrSeen(nodeTable->relid, currAttr->varno, currAttr->varattno, currAttr); //mappedTo here just means the attributeMap that has same table/column
			if(mappedTo != NULL) {
				char temp_args[100];

				currAttr->dimension = mappedTo->dimension;
				currAttr->depid = mappedTo->depid;
				currAttr->origdimnum = mappedTo->origdimnum;
				currAttr->mappedTo = mappedTo->mappedTo;

				currAttr->expr = (Expr *) makeConst(getUncertainOid(), (int32) -1, -1, (Datum) 0, true, false);

				strcpy(temp_args, mappedTo->mappedTo->marg_arg);

				strcat(mappedTo->mappedTo->resname, "_");
				strcat(mappedTo->mappedTo->resname, currAttr->resname);
				snprintf(mappedTo->mappedTo->marg_arg, 100, "%s,%d", temp_args, (currAttr->dimension-1));
//				continue;
			}
			else if (isJointlyDistributed(currAttr->resorigtbl, currAttr->varattno, &dimnum, &depid, &depdim)) {
				//check if this depid has been seen before
				mappedTo = depidSeen(depid, currAttr->varno);
				if(mappedTo != NULL) { //depid seen before  -- related to the same table no
					//change this attributes value to NULL
					char temp_args[100];

					currAttr->mappedTo = mappedTo;
					currAttr->expr = (Expr *) makeConst(getUncertainOid(), (int32) -1, -1, (Datum) 0, true, false);

					strcat(mappedTo->mappedTo->resname, "_");
					strcat(mappedTo->mappedTo->resname, currAttr->resname);
					strcpy(temp_args, mappedTo->marg_arg);
					snprintf(mappedTo->mappedTo->marg_arg, 100, "%s,%d", temp_args, (dimnum-1));

				} else {//depid not seen before
					AttrNumber jointAttrNum = getAttrNum(depid);

					currAttr->expr = (Expr *) makeVar(currAttr->varno, 
											jointAttrNum, 
											getVarType(currAttr->resorigtbl, jointAttrNum),  
											(int32) -1, 
											(Index) 0);
					snprintf(currAttr->marg_arg, 100, "%d", (dimnum-1));
				}
				currAttr->maxdim = depdim;
				currAttr->depid = depid;
				currAttr->origdimnum = dimnum;
				currAttr->dimension = dimnum;
			}
		}
	}

	//DEBUG("done with loop");
	endAttrMap = currAttr;
	//Take care of merge because of self join
	index = 0;
	for(currAttr = attrMap; currAttr != endAttrMap; currAttr = currAttr->next) {
		index++;
		mappedTo = attrSeenSelfJoin(currAttr->resorigtbl, currAttr->varno, currAttr->varattno, currAttr);
		if(mappedTo != NULL) {
			DEBUG("index : %d", index);
			mergeAttributeMap(mappedTo, currAttr);
		}
	}


	//printAttrMap();
	//Now from the where clause, pick the uncertain attributes and make corresponding AttributeMap
	if (qry->jointree->quals != NULL) {
		updateWhereAttrMap(qry, qry->jointree->quals, finalAttrMap);
		if(attrMap == NULL) {
//			DEBUG("attrMap is NULL");
			return;
		}
//		printAttrMap();
		//applyFloors(qry, qry->jointree->quals);
		//pfree(qry->jointree->quals);
		//qry->jointree->quals=NULL;
		qry->jointree->quals = applyFloors(qry, qry->jointree->quals);
	}
	DEBUG("createAttrMap Done");
}

void freeAttrMap() {
//	DEBUG("freeAttrMap");

	AttributeMap *curr=attrMap;
	AttributeMap *temp;

	while (curr!=NULL) {
		temp = curr->next;
		pfree(curr);
		curr=temp;
	}
	attrMap = NULL;
//	DEBUG("freeAttrMap Done");
}



/* Apply u_mass function -- basically you do a query over subQry with u_mass*/
Query *returnNonzeroTuples(Query *subQry, BoolExpr *newAND) {
//	DEBUG("returnNonzeroTuples");
	Query *qry;
	RangeTblEntry *newRTE;
	RangeTblRef *newRTR;
//	FuncExpr *newFE;
	Var *newV;
	List *newArgs;	
	TargetEntry *newTE, *attr;
	int i;
	ListCell *l;
	char *name;
	
	
	qry = makeNode(Query);
	qry->commandType = CMD_SELECT;
	qry->querySource = QSRC_ORIGINAL;
	qry->canSetTag = true;
	qry->utilityStmt = NULL;
	qry->resultRelation = 0;
	qry->intoClause = NULL;
	qry->hasAggs = false;
	qry->hasSubLinks = false;
	
	/* Make the original qury a subquery */
	newRTE = makeNode(RangeTblEntry);
	newRTE->alias = makeAlias("temp", NULL);
	
	newArgs = NULL;
	foreach(l, subQry->targetList) {
		attr = (TargetEntry *) lfirst(l);
		name = palloc(strlen(attr->resname) + 1);
		strcpy(name, attr->resname);
		newArgs = lappend(newArgs, makeString(name));
	}
		
	newRTE->eref = makeAlias("temp", newArgs);
	newRTE->rtekind = RTE_SUBQUERY;
	newRTE->subquery = subQry;
	newRTE->inh = false;
	newRTE->inFromCl = true;
	newRTE->requiredPerms = ((RangeTblEntry *)list_nth(subQry->rtable,0))->requiredPerms;
	newRTE->checkAsUser = ((RangeTblEntry *)list_nth(subQry->rtable,0))->checkAsUser;
	
	qry->rtable = lappend(NULL, newRTE);
	
	newRTR = makeNode(RangeTblRef);
	newRTR->rtindex = 1;

// 	/* make the quals node */
// 	newV = makeVar(1, whereAttrNo, getUncertainOid(), -1, 0);
// 	newFE = makeFuncExpr(getFuncOid("u_nonzero"), getTypeOid("bool"), lappend(NULL, newV), (CoercionForm) 0);

	qry->jointree = makeFromExpr(lappend(NULL,newRTR), (Node *)newAND);
	//qry->jointree->quals = newAND;

	/* Make the targetList Node */
	newArgs = NULL;
	i = 1;
	foreach(l, subQry->targetList) {
		attr = (TargetEntry *) lfirst(l);
		
		newV = (IsA(attr->expr, Var)? ((Var *)attr->expr) : NULL);		
		newV = makeVar(1, i, ((newV == NULL) ? getUncertainOid(): newV->vartype), -1, 0);
		
		newTE = makeTargetEntry((Expr *)newV, attr->resno, attr->resname, attr->resjunk);
		
		newArgs = lappend(newArgs, newTE);
		i++;
	}
	qry->targetList = newArgs;
	
	qry->returningList = NULL;
	qry->groupClause = NULL;
	qry->havingQual = NULL;
	qry->distinctClause = NULL;
	qry->sortClause = NULL;
	
	qry->limitOffset = subQry->limitOffset;
	qry->limitCount = subQry->limitCount;
	subQry->limitOffset = NULL;
	subQry->limitCount = NULL;
	
	qry->rowMarks = NULL;
	qry->setOperations = NULL;
	
//	DEBUG("returnNonzeroTuples finished");
	return qry;	
}


Query* updateTEs(Query *subQry) {
//	DEBUG("updateTEs");
	AttributeMap *temp;
	Const *cnode;
	char dim_list[100];
	BoolExpr *newAND;
	Var *newV;
	FuncExpr *newFE;
	List *newArgs = NULL;

	newAND = makeNode(BoolExpr);
	newAND->boolop = AND_EXPR;


	for(temp = attrMap; (temp != NULL)&&(temp->node != NULL); temp = temp->next) {
		if(temp->expr !=NULL) {
			pfree(temp->node->expr);
			temp->node->expr = temp->expr;
		}
		temp->node->ressortgroupref = 0;
		temp->node->resorigtbl = InvalidOid;
		temp->node->resorigcol = 0;

		if(temp->mappedTo == temp) {
			newArgs = NULL;
			newV = makeVar(1, temp->node->resno, getUncertainOid(), -1, 0);
			newFE = makeFuncExpr(getFuncOid("u_nonzero"), getTypeOid("bool"), lappend(NULL, newV), (CoercionForm) 0);
			newAND->args = lappend(newAND->args, newFE);

			strcat(temp->node->resname, "_");
			strcat(temp->node->resname, temp->resname);
			snprintf(dim_list, 100, "{%s}", temp->marg_arg);
			cnode = makeConst(get_int2Oid(), (int32) -1, -1,  
						OidFunctionCall3(F_ARRAY_IN, CStringGetDatum(dim_list), 
						ObjectIdGetDatum(INT2OID), -1), 
						false, false);

			newArgs = lappend(newArgs, temp->node->expr);
			newArgs = lappend(newArgs, cnode);		

			temp->node->expr = (Expr *)makeFuncExpr(getU_marginOid(), 
								getUncertainOid(), 
								newArgs, 
								COERCE_EXPLICIT_CALL);
		}
	}
//	DEBUG("updateTEs finished");
//	return subQry;
	return returnNonzeroTuples(subQry, newAND);
}

bool hasUncertain(List *rtablelist) {
	RangeTblEntry *node;
	ListCell *l;
//	Index tableIndex = 0;
	int i;
	Oid typeOid;
	char sql[200];
	int32 ret;
	bool uncertainExists = false;
	bool isnull;
	
	isUncertainType((Oid) 0); /* Pre-cache to make sure that the SPI calls are not nested */
	
	/* connect to the DB */
	SPI_connect();
	
	foreach(l, rtablelist) {
		node = (RangeTblEntry *) lfirst(l);

		if (node->rtekind == RTE_RELATION) {
			snprintf(sql, 200, "SELECT atttypid, attnum FROM pg_attribute WHERE attrelid = %d and attnum > 0 ORDER by attnum", node->relid);

			ret = SPI_execute(sql, true, 0);
			if (ret == SPI_OK_SELECT) {
				TupleDesc tupdesc = SPI_tuptable->tupdesc;
			
				for(i = 0; i < SPI_processed; i++) {
					HeapTuple tuple = SPI_tuptable->vals[i];
					typeOid = DatumGetObjectId(SPI_getbinval(tuple, tupdesc, 1, &isnull));
					if (isUncertainType(typeOid)) { /* Process this attribute */
						uncertainExists = true;
						break;
					}
				}
			} 
		}
		if(uncertainExists)
			break;
	}
	/* close the DB connection */
	SPI_finish();
	return uncertainExists;
}


Query *transformOrionSelectStmt(Query *qry) {
//	DEBUG("transformOrionSelectStmt");

	if (ActiveSnapshot == NULL)
		ActiveSnapshot = GetTransactionSnapshot();
	else
		return qry;

	
	if(! hasUncertain(qry->rtable)) {
		ActiveSnapshot = NULL;
//		DEBUG("transformOrionSelectStmt finished as there is no uncertain");
		return qry;
	}
//	orion_debug_query_pprint("Original Query :", qry);

	if(qry->jointree->quals != NULL) {
		qry->jointree->quals = convertToCNF(qry->jointree->quals);
//		orion_debug_query_pprint("after convertToCNF Query :", qry);
	}


	/* Create the attrmap */
	createAttrMap(qry);
	/* Take care of self join */
	//mergeSelfJoin(qry);
	qry = updateTEs(qry);
	freeAttrMap();

//	orion_debug_query_pprint("Final Query :", qry);
//	DEBUG("transformOrionSelectStmt finished");

	ActiveSnapshot = NULL;
	return qry;
}
